﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Reflection;
using System.IO;
using UnityEngine;

namespace TreeEdit
{
    [KSPAddon(KSPAddon.Startup.Instantly, false)]
    public class TreeLoaderSpawner : MonoBehaviour
    {
        void Start()
        {
            GameObject manager = (GameObject)GameObject.Find("TreeLoaderManager");
            if (manager == null)
            {
                manager = new GameObject("TreeLoaderManager", typeof(TreeLoader));
                GameObject.DontDestroyOnLoad(manager);
                Destroy(this);
            }
            else
            {
                manager.SendMessage("GetVersion", this, SendMessageOptions.DontRequireReceiver);
            }
        }

        public void VersionCallback(System.Version version)
        {
            System.Version mine = Assembly.GetAssembly(this.GetType()).GetName().Version;
            if (mine > version)
            {
                Destroy(GameObject.Find("TreeLoaderManager"));
                GameObject manager = new GameObject("TreeLoaderManager", typeof(TreeLoader));
                GameObject.DontDestroyOnLoad(manager);
            }
            Destroy(this);
        }
    }

    public class TreeLoader : MonoBehaviour
    {
        const string server = "http://www.kspmodders.com/treeedit/backend.php";

        static bool loadNodes = false, refreshNodes = false;
        static Dictionary<string, string> originalPartTech;

        GUIContent[] onlineList;
        List<TreeListItemData> onlineItems = new List<TreeListItemData>();
        ComboBox onlineCombo;
        GUIStyle listStyle = null;
        WWW loader;
        bool loadingList = false, loadingFile = false, showTreeSelectPopup = false;


        public void GetVersion(object caller)
        {
            caller.GetType().GetMethod("VersionCallback").Invoke(caller, new object[] { Assembly.GetAssembly(this.GetType()).GetName().Version });
        }

        void Start()
        {
            GameEvents.onGUIRnDComplexSpawn.Add(new EventVoid.OnEvent(OnGUIRnDComplexSpawn));
            GameEvents.onGameSceneLoadRequested.Add(new EventData<GameScenes>.OnEvent(OnGameSceneLoadRequested));
            GameEvents.OnTechnologyResearched.Add(new EventData<GameEvents.HostTargetAction<RDTech, RDTech.OperationResult>>.OnEvent(OnTechnologyResearched));

            onlineList = new GUIContent[] { new GUIContent("Loading...") };
            onlineCombo = new ComboBox(new Rect(400, 2, 200, 20), onlineList[0], onlineList, null);
            onlineCombo.lineHeight = 25;
        }

        void OnDestroy()
        {
            GameEvents.onGUIRnDComplexSpawn.Remove(new EventVoid.OnEvent(OnGUIRnDComplexSpawn));
            GameEvents.onGameSceneLoadRequested.Remove(new EventData<GameScenes>.OnEvent(OnGameSceneLoadRequested));
            GameEvents.OnTechnologyResearched.Remove(new EventData<GameEvents.HostTargetAction<RDTech, RDTech.OperationResult>>.OnEvent(OnTechnologyResearched));
        }

        void OnGUIRnDComplexSpawn()
        {
            loadNodes = true;
        }

        void OnTechnologyResearched(GameEvents.HostTargetAction<RDTech, RDTech.OperationResult> evt)
        {
            RedrawTreeNodes();
        }

        void OnGameSceneLoadRequested(GameScenes scene)
        {
            if (scene == GameScenes.MAINMENU && originalPartTech == null)
            {
                originalPartTech = new Dictionary<string, string>();
                foreach (AvailablePart p in PartLoader.LoadedPartsList)
                {
                    if (!originalPartTech.ContainsKey(p.name))
                    {
                        originalPartTech.Add(p.name, p.TechRequired);
                    }
                    else
                    {
                        print("Treeloader - found duplicate part name: " + p.name);
                    }
                }
            }
            if (scene == GameScenes.SPACECENTER && HighLogic.CurrentGame.Mode == Game.Modes.CAREER)
            {
                if (File.Exists(KSPUtil.ApplicationRootPath + "/saves/" + HighLogic.SaveFolder + "/tree.cfg"))
                {
                    print("TreeLoader version " + Assembly.GetAssembly(this.GetType()).GetName().Version.ToString(3) + " loading parts...");
                    LoadParts(KSPUtil.ApplicationRootPath + "/saves/" + HighLogic.SaveFolder + "/tree.cfg");
                }
                else if (!File.Exists(KSPUtil.ApplicationRootPath + "/saves/" + HighLogic.SaveFolder + "/notree.cfg"))
                {
                    showTreeSelectPopup = true;
                    ReloadList();
                }
                else
                {
                    print("TreeLoader version " + Assembly.GetAssembly(this.GetType()).GetName().Version.ToString(3) + ", using stock settings...");
                    LoadStockParts();
                }
            }
        }

        void ReloadList()
        {
            loader = new WWW(server + "?op=treelist");
            loadingList = true;
            onlineList = new GUIContent[] { new GUIContent("Loading...") };
            onlineCombo.buttonContent = onlineList[0];
            onlineCombo.selectedItemIndex = 0;
        }

        void Update()
        {
            if (refreshNodes || (Input.GetKeyDown(KeyCode.F7) && !Input.GetKey(KeyCode.LeftAlt)))
            {
                refreshNodes = false;
                RedrawTreeNodes();
            }
            if (loadNodes || (Input.GetKeyDown(KeyCode.F7) && Input.GetKey(KeyCode.LeftAlt)))
            {
                loadNodes = false;
                if (File.Exists(KSPUtil.ApplicationRootPath + "/saves/" + HighLogic.SaveFolder + "/tree.cfg"))
                {
                    LoadNodes(KSPUtil.ApplicationRootPath + "/saves/" + HighLogic.SaveFolder + "/tree.cfg");
                }
            }
            if (loadingList && loader.isDone)
            {
                string[] lines = loader.text.Split(new char[] { '\n', '\r' }, StringSplitOptions.RemoveEmptyEntries);
                onlineItems.Clear();
                onlineList = new GUIContent[lines.Length + 2];
                onlineList[0] = new GUIContent("-- Select --");
                onlineList[lines.Length + 1] = new GUIContent("-- Reload List --");

                for (int i = 0; i < lines.Length; i++)
                {
                    string[] pcs = lines[i].Split('|');
                    if (pcs.Length == 5)
                    {
                        onlineList[i + 1] = new GUIContent(pcs[1]);
                        onlineItems.Add(new TreeListItemData(int.Parse(pcs[0]), pcs[1], pcs[2] == "1", pcs[3] == "1", pcs[4] == "1"));
                    }
                }
                onlineCombo.listContent = onlineList;
                onlineCombo.buttonContent = onlineList[0];
                onlineCombo.selectedItemIndex = 0;

                loadingList = false;
            }
            if (loadingFile && loader.isDone)
            {
                File.WriteAllText(KSPUtil.ApplicationRootPath + "/saves/" + HighLogic.SaveFolder + "/tree.cfg", loader.text);
                try
                {
                    LoadParts(KSPUtil.ApplicationRootPath + "/saves/" + HighLogic.SaveFolder + "/tree.cfg");
                }
                catch (Exception)
                {
                }
                loadingFile = false;
            }
            if (onlineCombo.selectedItemIndex > onlineItems.Count)
            {
                ReloadList();
            }
        }

        Rect cboPos = new Rect();

        void OnGUI()
        {
            if (showTreeSelectPopup && HighLogic.LoadedScene == GameScenes.SPACECENTER)
            {
                GUI.skin = AssetBase.GetGUISkin("KSP window 2");

                if (listStyle == null)
                {
                    listStyle = new GUIStyle("label");
                    listStyle.padding.left = listStyle.padding.right = 5;
                    onlineCombo.listStyle = listStyle;
                }

                GUILayout.BeginArea(new Rect(Screen.width / 2 - 250, Screen.height / 2 - 30, 500, 60), "MuMech TreeLoader", GUI.skin.window);
                GUILayout.BeginHorizontal();
                GUILayout.Label("Online tree:");
                Rect cbo = GUILayoutUtility.GetRect(275, 25);
                if (Event.current.type == EventType.Repaint)
                {
                    cboPos = cbo;
                }
                if (onlineCombo.selectedItemIndex > 0 && onlineCombo.selectedItemIndex <= onlineItems.Count)
                {
                    if (GUILayout.Button("Load", GUILayout.Height(20)))
                    {
                        loader = new WWW(server + "?op=treeload&id=" + onlineItems[onlineCombo.selectedItemIndex - 1].id);
                        loadingFile = true;
                        showTreeSelectPopup = false;
                    }
                }
                GUILayout.FlexibleSpace();
                if (GUILayout.Button("Use Stock", GUILayout.Height(20)))
                {
                    File.WriteAllText(KSPUtil.ApplicationRootPath + "/saves/" + HighLogic.SaveFolder + "/notree.cfg", "");
                    LoadStockParts();
                    showTreeSelectPopup = false;
                }
                GUILayout.EndHorizontal();
                GUILayout.EndArea();

                onlineCombo.rect = new Rect(cboPos.xMin + Screen.width / 2 - 250, cboPos.yMin + Screen.height / 2 - 26, 275, 20);
                onlineCombo.Show();
            }
        }

        public static void RedrawTreeNodes()
        {
            foreach (RDNode n in GameObject.FindObjectsOfType(typeof(RDNode)).Cast<RDNode>().Where(a => a.treeNode))
            {
                for (int i = 0; i < n.parents.Length; i++)
                {
                    if (n.parents[i].line != null)
                    {
                        Vector.DestroyLine(ref n.parents[i].line);
                    }
                    if (n.parents[i].arrowHead != null)
                    {
                        GameObject.Destroy(n.parents[i].arrowHead);
                    }

                    Vector3 pos = n.transform.position - n.parents[i].parent.node.transform.position;
                    if (Mathf.Abs(pos.x) > Mathf.Abs(pos.y))
                    {
                        if (pos.x > 0)
                        {
                            n.parents[i].anchor = RDNode.Anchor.LEFT;
                            n.parents[i].parent.anchor = RDNode.Anchor.RIGHT;
                        }
                        else
                        {
                            n.parents[i].anchor = RDNode.Anchor.RIGHT;
                            n.parents[i].parent.anchor = RDNode.Anchor.LEFT;
                        }
                    }
                    else
                    {
                        if (pos.y > 0)
                        {
                            n.parents[i].anchor = RDNode.Anchor.BOTTOM;
                            n.parents[i].parent.anchor = RDNode.Anchor.TOP;
                        }
                        else
                        {
                            if (pos.x > 0)
                            {
                                n.parents[i].anchor = RDNode.Anchor.LEFT;
                                n.parents[i].parent.anchor = RDNode.Anchor.RIGHT;
                            }
                            else
                            {
                                n.parents[i].anchor = RDNode.Anchor.RIGHT;
                                n.parents[i].parent.anchor = RDNode.Anchor.LEFT;
                            }
                        }
                    }
                }
                if (n != null && (!n.hideIfNoParts || n.PartsInTotal() > 0))
                {
                    if (n.controller == null)
                    {
                        n.controller = (RDController)GameObject.FindObjectOfType(typeof(RDController));
                    }
                    if (n.parents.Length > 0 && n.parents.Count(a => a.parent.node.IsResearched) == n.parents.Length)
                    {
                        typeof(RDNode).GetMethod("DrawArrow", BindingFlags.NonPublic | BindingFlags.Instance).Invoke(n, new object[] { n.controller.gridArea.LineMaterial });
                    }
                    else if (n.parents.Count(a => a.parent.node.IsResearched) > 0)
                    {
                        typeof(RDNode).GetMethod("DrawArrow", BindingFlags.NonPublic | BindingFlags.Instance).Invoke(n, new object[] { n.controller.gridArea.LineMaterialGray });
                    }
                }
                try
                {
                    n.UpdateGraphics();
                }
                catch (Exception)
                {
                }
            }
        }

        static GameObject nodePrefab;
        static RDNode model;

        public static List<RDNode> removedNodes = new List<RDNode>();

        public static RDNode AddNode()
        {
            if (model == null || model.controller == null)
            {
                model = (from a in GameObject.FindObjectsOfType(typeof(RDNode)) where ((RDNode)a).treeNode && ((RDNode)a).gameObject.activeSelf select (RDNode)a).First();
                if (nodePrefab != null && model.controller != null)
                {
                    Destroy(nodePrefab);
                    nodePrefab = null;
                }
            }
            if (model.controller != null && model.controller.node_selected != null)
            {
                model = model.controller.node_selected;
                if (nodePrefab != null)
                {
                    Destroy(nodePrefab);
                    nodePrefab = null;
                }
            }
            if (nodePrefab == null)
            {
                nodePrefab = new GameObject("newnode", typeof(RDNode), typeof(RDTech));
                nodePrefab.GetComponent<RDTech>().techID = "newTech_RenameMe";
                nodePrefab.GetComponent<RDNode>().prefab = model.prefab;
                nodePrefab.GetComponent<RDNode>().parents = new RDNode.Parent[0];
                nodePrefab.GetComponent<RDNode>().icon = model.icon;
                nodePrefab.GetComponent<RDNode>().controller = model.controller;
                nodePrefab.GetComponent<RDNode>().scale = model.scale;
                nodePrefab.SetActive(false);
            }
            GameObject clone = (GameObject)GameObject.Instantiate(nodePrefab);
            clone.SetActive(true);
            clone.transform.parent = model.transform.parent;
            clone.transform.localPosition = model.transform.localPosition + new Vector3(0, 50, 0);
            clone.GetComponent<RDNode>().children = new List<RDNode>();
            return clone.GetComponent<RDNode>();
        }

        public static void RemoveNode(RDNode n)
        {
            if (n.gameObject.name.StartsWith("newnode_"))
            {
                foreach (RDNode c in n.children)
                {
                    int index = (from i in Enumerable.Range(0, c.parents.Length) where c.parents[i].parent.node == n select i).First();

                    if (c.parents[index].line != null)
                    {
                        Vector.DestroyLine(ref c.parents[index].line);
                    }
                    if (c.parents[index].arrowHead != null)
                    {
                        GameObject.Destroy(c.parents[index].arrowHead);
                    }

                    List<RDNode.Parent> tmp = c.parents.ToList();
                    tmp.RemoveAt(index);
                    c.parents = tmp.ToArray();
                }

                foreach (RDNode.Parent p in n.parents)
                {
                    if (p.line != null)
                    {
                        Vector.DestroyLine(ref p.line);
                    }
                    if (p.arrowHead != null)
                    {
                        GameObject.Destroy(p.arrowHead);
                    }
                    if (p.parent.node.children.Contains(n))
                    {
                        p.parent.node.children.Remove(n);
                    }
                }

                DestroyImmediate(n.gameObject);
            }
            else
            {
                removedNodes.Add(n);

                foreach (RDNode c in n.children)
                {
                    int index = (from i in Enumerable.Range(0, c.parents.Length) where c.parents[i].parent.node == n select i).First();

                    if (c.parents[index].line != null)
                    {
                        Vector.DestroyLine(ref c.parents[index].line);
                    }
                    if (c.parents[index].arrowHead != null)
                    {
                        GameObject.Destroy(c.parents[index].arrowHead);
                    }

                    List<RDNode.Parent> tmp = c.parents.ToList();
                    tmp.RemoveAt(index);
                    c.parents = tmp.ToArray();
                }
                n.children = new List<RDNode>();

                foreach (RDNode.Parent p in n.parents)
                {
                    if (p.line != null)
                    {
                        Vector.DestroyLine(ref p.line);
                    }
                    if (p.arrowHead != null)
                    {
                        GameObject.Destroy(p.arrowHead);
                    }
                    if (p.parent.node.children.Contains(n))
                    {
                        p.parent.node.children.Remove(n);
                    }
                }
                n.parents = new RDNode.Parent[0];

                n.gameObject.SetActive(false);
            }
        }

        public static void LoadNodes(string file)
        {
            RDController controller = (RDController)GameObject.FindObjectOfType(typeof(RDController));

            if (controller != null && controller.gridArea != null)
            {
                typeof(RDGridArea).GetMethod("ZoomTo", BindingFlags.NonPublic | BindingFlags.Instance).Invoke(controller.gridArea, new object[] { 1f, true });
            }

            if (File.Exists(file))
            {
                Dictionary<string, RDNode> name2node = new Dictionary<string, RDNode>();

                foreach (RDNode n in removedNodes)
                {
                    if (n != null && n.gameObject != null)
                    {
                        n.gameObject.SetActive(true);
                    }
                }

                foreach (RDNode n in GameObject.FindObjectsOfType(typeof(RDNode)).Cast<RDNode>().Where(a => a.treeNode))
                {
                    if (n.gameObject.name.StartsWith("newnode_"))
                    {
                        RemoveNode(n);
                    }
                    else if (!name2node.ContainsKey(n.gameObject.name))
                    {
                        foreach (RDNode.Parent p in n.parents)
                        {
                            if (p.line != null)
                            {
                                Vector.DestroyLine(ref p.line);
                            }
                            if (p.arrowHead != null)
                            {
                                GameObject.Destroy(p.arrowHead);
                            }
                            p.parent.node.children = new List<RDNode>();
                        }
                        n.parents = new RDNode.Parent[0];

                        name2node.Add(n.gameObject.name, n);
                    }
                    else
                    {
                        print("Duplicate part name: " + n.gameObject.name);
                    }
                }

                try
                {
                    ConfigNode cfg = ConfigNode.Load(file);
                    removedNodes.Clear();
                    foreach (ConfigNode node in cfg.GetNodes("REMOVENODE"))
                    {
                        string name = node.GetValue("name");
                        GameObject badNode = GameObject.Find(name);
                        if (badNode != null)
                        {
                            RemoveNode(badNode.GetComponent<RDNode>());
                        }
                    }
                    List<string> doneNodes = new List<string>();
                    Dictionary<string, string> parentList = new Dictionary<string, string>();

                    foreach (ConfigNode node in cfg.GetNodes("NODE"))
                    {
                        string name = node.GetValue("name");

                        if (String.IsNullOrEmpty(name))
                        {
                            continue;
                        }
                        if (!name2node.ContainsKey(name))
                        {
                            if (!name.StartsWith("newnode_"))
                            {
                                name = "newnode_" + name;
                            }
                            if (!name2node.ContainsKey(name))
                            {
                                print("Adding new node: " + name);
                                RDNode newNode = AddNode();
                                newNode.gameObject.name = name;
                                name2node.Add(name, newNode);
                            }
                        }

                        if (node.HasValue("pos"))
                        {
                            name2node[name].transform.localPosition = ConfigNode.ParseVector3(node.GetValue("pos"));
                        }
                        if (node.HasValue("icon"))
                        {
                            name2node[name].icon = (RDNode.Icon)ConfigNode.ParseEnum(typeof(RDNode.Icon), node.GetValue("icon"));
                        }
                        if (node.HasValue("techID"))
                        {
                            name2node[name].tech.techID = node.GetValue("techID");
                        }
                        if (node.HasValue("cost"))
                        {
                            name2node[name].tech.scienceCost = int.Parse(node.GetValue("cost"));
                        }
                        if (node.HasValue("title"))
                        {
                            name2node[name].tech.title = name2node[name].name = node.GetValue("title");
                        }
                        if (node.HasValue("description"))
                        {
                            name2node[name].tech.description = name2node[name].description = node.GetValue("description");
                        }
                        if (node.HasValue("anyParent"))
                        {
                            name2node[name].AnyParentToUnlock = bool.Parse(node.GetValue("anyParent"));
                        }
                        if (node.HasValue("hideIfEmpty"))
                        {
                            name2node[name].tech.hideIfNoParts = bool.Parse(node.GetValue("hideIfEmpty"));
                        }

                        name2node[name].gameObject.SetActive(true);

                        if (node.HasValue("parents"))
                        {
                            parentList.Add(name, node.GetValue("parents"));
                        }

                        if (ResearchAndDevelopment.Instance != null)
                        {
                            ProtoTechNode tn = ResearchAndDevelopment.Instance.GetTechState(name2node[name].tech.techID);
                            if (tn == null)
                            {
                                tn = new ProtoTechNode();
                                tn.partsPurchased = new List<AvailablePart>();
                                ResearchAndDevelopment.Instance.SetTechState(name2node[name].tech.techID, tn);
                            }

                            name2node[name].tech.state = tn.state;
                            name2node[name].tech.partsPurchased = tn.partsPurchased;

                            if (node.HasNode("PARTS"))
                            {
                                name2node[name].tech.partsAssigned = PartLoader.LoadedPartsList.Where(a => node.GetNode("PARTS").GetValues("name").Any(x => x == a.name) || a.TechRequired == name2node[name].tech.techID).ToList();
                                name2node[name].tech.partsPurchased.RemoveAll(a => !name2node[name].tech.partsAssigned.Contains(a));
                            }

                            tn.UpdateFromTechNode(name2node[name].tech);
                        }

                        doneNodes.Add(name);
                    }

                    foreach (string name in parentList.Keys)
                    {
                        List<RDNode.Parent> parents = new List<RDNode.Parent>();
                        string[] newParents = parentList[name].Split(new char[] { ',', ' ', '\t', '\r', '\n' }, StringSplitOptions.RemoveEmptyEntries);
                        foreach (string p in newParents)
                        {
                            if (name2node.ContainsKey(p))
                            {
                                parents.Add(new RDNode.Parent(new RDNode.ParentAnchor(name2node[p], RDNode.Anchor.RIGHT), RDNode.Anchor.LEFT));
                                name2node[p].children.Add(name2node[name]);
                            }
                        }
                        name2node[name].parents = parents.ToArray();
                        name2node[name].SetIconState(name2node[name].icon);
                    }

                    foreach (string name in name2node.Keys.Where(a => !doneNodes.Contains(a)))
                    {
                        print("Auto-removing node: " + name);
                        RemoveNode(name2node[name]);
                    }
                }
                catch (Exception e)
                {
                    print("TreeLoader.LoadNodes: Failed to load tree - " + e.Message + "\n" + e.StackTrace);
                }
                refreshNodes = true;
            }
        }

        public static void LoadParts(string file)
        {
            if (File.Exists(file) && PartLoader.Instance != null && PartLoader.Instance.IsReady())
            {
                Dictionary<string, string> part2node = new Dictionary<string, string>();

                try
                {
                    ConfigNode cfg = ConfigNode.Load(file);
                    foreach (ConfigNode node in cfg.GetNodes("NODE"))
                    {
                        string id = node.GetValue("techID");

                        if (node.HasNode("PARTS"))
                        {
                            foreach (string p in node.GetNode("PARTS").GetValues("name"))
                            {
                                if (part2node.ContainsKey(p))
                                {
                                    print("TreeLoader: Part in multiple nodes:" + p + " - " + part2node[p] + ", " + id);
                                }
                                else
                                {
                                    part2node.Add(p, id);
                                }
                            }
                        }
                    }

                    foreach (AvailablePart a in PartLoader.LoadedPartsList)
                    {
                        if (part2node.ContainsKey(a.name.Trim()))
                        {
                            a.TechRequired = part2node[a.name];
                        }
                    }
                }
                catch (Exception e)
                {
                    print("TreeLoader.LoadParts: Failed to load tree - " + e.Message + "\n" + e.StackTrace);
                }
            }
        }

        public static void LoadStockParts()
        {
            if (originalPartTech != null)
            {
                foreach (AvailablePart p in PartLoader.LoadedPartsList)
                {
                    if (originalPartTech.ContainsKey(p.name))
                    {
                        p.TechRequired = originalPartTech[p.name];
                    }
                }
            }
        }
    }
}
